@@ -38,46 +38,19 @@
 #include <libssh/string.h>
 #include <libssh/server.h>
 
-/** current state of an GSSAPI authentication */
-enum ssh_gssapi_state_e {
-    SSH_GSSAPI_STATE_NONE, /* no status */
-    SSH_GSSAPI_STATE_RCV_TOKEN, /* Expecting a token */
-    SSH_GSSAPI_STATE_RCV_MIC, /* Expecting a MIC */
-};
-
-struct ssh_gssapi_struct{
-    enum ssh_gssapi_state_e state; /* current state */
-    struct gss_OID_desc_struct mech; /* mechanism being elected for auth */
-    gss_cred_id_t server_creds; /* credentials of server */
-    gss_cred_id_t client_creds; /* creds delegated by the client */
-    gss_ctx_id_t ctx; /* the authentication context */
-    gss_name_t client_name; /* Identity of the client */
-    char *user; /* username of client */
-    char *canonic_user; /* canonic form of the client's username */
-    char *service; /* name of the service */
-    struct {
-        gss_name_t server_name; /* identity of server */
-        OM_uint32 flags; /* flags used for init context */
-        gss_OID oid; /* mech being used for authentication */
-        gss_cred_id_t creds; /* creds used to initialize context */
-        gss_cred_id_t client_deleg_creds; /* delegated creds (const, not freeable) */
-    } client;
-};
-
-
 /** @internal
  * @initializes a gssapi context for authentication
  */
-static int ssh_gssapi_init(ssh_session session)
+int
+ssh_gssapi_init(ssh_session session)
 {
     if (session->gssapi != NULL)
         return SSH_OK;
-    session->gssapi = malloc(sizeof(struct ssh_gssapi_struct));
-    if(!session->gssapi){
+    session->gssapi = calloc(1, sizeof(struct ssh_gssapi_struct));
+    if (session->gssapi == NULL) {
         ssh_set_error_oom(session);
         return SSH_ERROR;
     }
-    ZERO_STRUCTP(session->gssapi);
     session->gssapi->server_creds = GSS_C_NO_CREDENTIAL;
     session->gssapi->client_creds = GSS_C_NO_CREDENTIAL;
     session->gssapi->ctx = GSS_C_NO_CONTEXT;
@@ -85,21 +58,60 @@ static int ssh_gssapi_init(ssh_session session)
     return SSH_OK;
 }
 
+void
+ssh_gssapi_log_error(int verb, const char *msg_a, int maj_stat, int min_stat)
+{
+    gss_buffer_desc msg = GSS_C_EMPTY_BUFFER;
+    OM_uint32 dummy_min;
+    OM_uint32 message_context = 0;
+
+    do {
+        gss_display_status(&dummy_min,
+                           maj_stat,
+                           GSS_C_GSS_CODE,
+                           GSS_C_NO_OID,
+                           &message_context,
+                           &msg);
+        SSH_LOG(verb, "GSSAPI(%s): %s", msg_a, (const char *)msg.value);
+        gss_release_buffer(&dummy_min, &msg);
+
+    } while (message_context != 0);
+
+    do {
+        gss_display_status(&dummy_min,
+                           min_stat,
+                           GSS_C_MECH_CODE,
+                           GSS_C_NO_OID,
+                           &message_context,
+                           &msg);
+        SSH_LOG(verb, "GSSAPI(%s): %s", msg_a, (const char *)msg.value);
+        gss_release_buffer(&dummy_min, &msg);
+
+    } while (message_context != 0);
+}
+
 /** @internal
  * @frees a gssapi context
  */
-static void ssh_gssapi_free(ssh_session session)
+void
+ssh_gssapi_free(ssh_session session)
 {
     OM_uint32 min;
     if (session->gssapi == NULL)
         return;
     SAFE_FREE(session->gssapi->user);
-    SAFE_FREE(session->gssapi->mech.elements);
+
+    gss_release_name(&min, &session->gssapi->client.server_name);
     gss_release_cred(&min,&session->gssapi->server_creds);
     if (session->gssapi->client.creds !=
                     session->gssapi->client.client_deleg_creds) {
         gss_release_cred(&min, &session->gssapi->client.creds);
     }
+    gss_release_oid(&min, &session->gssapi->client.oid);
+    gss_delete_sec_context(&min, &session->gssapi->ctx, GSS_C_NO_BUFFER);
+
+    SAFE_FREE(session->gssapi->mech.elements);
+    SAFE_FREE(session->gssapi->canonic_user);
     SAFE_FREE(session->gssapi);
 }
 
@@ -133,55 +145,6 @@ static int ssh_gssapi_send_response(ssh_session session, ssh_string oid)
 
 #endif /* WITH_SERVER */
 
-static void ssh_gssapi_log_error(int verb,
-                                 const char *msg,
-                                 int maj_stat,
-                                 int min_stat)
-{
-    gss_buffer_desc msg_maj = {
-        .length = 0,
-    };
-    gss_buffer_desc msg_min = {
-        .length = 0,
-    };
-    OM_uint32 dummy_maj, dummy_min;
-    OM_uint32 message_context = 0;
-
-    dummy_maj = gss_display_status(&dummy_min,
-                                   maj_stat,
-                                   GSS_C_GSS_CODE,
-                                   GSS_C_NO_OID,
-                                   &message_context,
-                                   &msg_maj);
-    if (dummy_maj != 0) {
-        goto out;
-    }
-
-    dummy_maj = gss_display_status(&dummy_min,
-                                   min_stat,
-                                   GSS_C_MECH_CODE,
-                                   GSS_C_NO_OID,
-                                   &message_context,
-                                   &msg_min);
-    if (dummy_maj != 0) {
-        goto out;
-    }
-
-    SSH_LOG(verb,
-            "GSSAPI(%s): %s - %s",
-            msg,
-            (const char *)msg_maj.value,
-            (const char *)msg_min.value);
-
-out:
-    if (msg_maj.value) {
-        gss_release_buffer(&dummy_min, &msg_maj);
-    }
-    if (msg_min.value) {
-        gss_release_buffer(&dummy_min, &msg_min);
-    }
-}
-
 #ifdef WITH_SERVER
 
 /** @internal
@@ -205,30 +168,43 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
     struct gss_OID_desc_struct oid;
     int rc;
 
-    if (ssh_callbacks_exists(session->server_callbacks, gssapi_select_oid_function)){
-        ssh_string oid_s = session->server_callbacks->gssapi_select_oid_function(session,
-                user, n_oid, oids,
-                session->server_callbacks->userdata);
-        if (oid_s != NULL){
-            if (ssh_gssapi_init(session) == SSH_ERROR)
-                return SSH_ERROR;
-            session->gssapi->state = SSH_GSSAPI_STATE_RCV_TOKEN;
+    /* Destroy earlier GSSAPI context if any */
+    ssh_gssapi_free(session);
+    rc = ssh_gssapi_init(session);
+    if (rc == SSH_ERROR)
+        return rc;
+
+    /* Callback should select oid and acquire credential */
+    if (ssh_callbacks_exists(session->server_callbacks,
+                             gssapi_select_oid_function)) {
+        ssh_string oid_s = NULL;
+        session->gssapi->state = SSH_GSSAPI_STATE_RCV_TOKEN;
+        SAFE_FREE(session->gssapi->user);
+        session->gssapi->user = strdup(user);
+        oid_s = session->server_callbacks->gssapi_select_oid_function(
+            session,
+            user,
+            n_oid,
+            oids,
+            session->server_callbacks->userdata);
+        if (oid_s != NULL) {
             rc = ssh_gssapi_send_response(session, oid_s);
-            SSH_STRING_FREE(oid_s);
             return rc;
         } else {
-            return ssh_auth_reply_default(session,0);
+            return ssh_auth_reply_default(session, 0);
         }
     }
+    /* Default implementation for selecting oid and acquiring credential */
     gss_create_empty_oid_set(&min_stat, &both_supported);
 
     maj_stat = gss_indicate_mechs(&min_stat, &supported);
     if (maj_stat != GSS_S_COMPLETE) {
-        SSH_LOG(SSH_LOG_WARNING, "indicate mecks %d, %d", maj_stat, min_stat);
-        ssh_gssapi_log_error(SSH_LOG_WARNING,
+        SSH_LOG(SSH_LOG_DEBUG, "indicate mechs %d, %d", maj_stat, min_stat);
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "indicate mechs",
                              maj_stat,
                              min_stat);
+        gss_release_oid_set(&min_stat, &both_supported);
         return SSH_ERROR;
     }
 
@@ -246,7 +222,7 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
             continue;
         }
         if(len < 2 || oid_s[0] != SSH_OID_TAG || ((size_t)oid_s[1]) != len - 2){
-            SSH_LOG(SSH_LOG_WARNING,"GSSAPI: received invalid OID");
+            SSH_LOG(SSH_LOG_TRACE,"GSSAPI: received invalid OID");
             continue;
         }
         oid.elements = &oid_s[2];
@@ -259,25 +235,23 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
     }
     gss_release_oid_set(&min_stat, &supported);
     if (oid_count == 0){
-        SSH_LOG(SSH_LOG_PROTOCOL,"GSSAPI: no OID match");
+        SSH_LOG(SSH_LOG_DEBUG,"GSSAPI: no OID match");
         ssh_auth_reply_default(session, 0);
         gss_release_oid_set(&min_stat, &both_supported);
         return SSH_OK;
     }
-    /* from now we have room for context */
-    if (ssh_gssapi_init(session) == SSH_ERROR)
-        return SSH_ERROR;
 
     name_buf.value = service_name;
     name_buf.length = strlen(name_buf.value) + 1;
     maj_stat = gss_import_name(&min_stat, &name_buf,
             (gss_OID) GSS_C_NT_HOSTBASED_SERVICE, &server_name);
     if (maj_stat != GSS_S_COMPLETE) {
-        SSH_LOG(SSH_LOG_WARNING, "importing name %d, %d", maj_stat, min_stat);
-        ssh_gssapi_log_error(SSH_LOG_WARNING,
+        SSH_LOG(SSH_LOG_DEBUG, "importing name %d, %d", maj_stat, min_stat);
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "importing name",
                              maj_stat,
                              min_stat);
+        gss_release_oid_set(&min_stat, &both_supported);
         return -1;
     }
 
@@ -288,8 +262,8 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
     gss_release_oid_set(&min_stat, &both_supported);
 
     if (maj_stat != GSS_S_COMPLETE) {
-        SSH_LOG(SSH_LOG_WARNING, "error acquiring credentials %d, %d", maj_stat, min_stat);
-        ssh_gssapi_log_error(SSH_LOG_WARNING,
+        SSH_LOG(SSH_LOG_TRACE, "error acquiring credentials %d, %d", maj_stat, min_stat);
+        ssh_gssapi_log_error(SSH_LOG_TRACE,
                              "acquiring creds",
                              maj_stat,
                              min_stat);
@@ -297,7 +271,7 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
         return SSH_ERROR;
     }
 
-    SSH_LOG(SSH_LOG_PROTOCOL, "acquiring credentials %d, %d", maj_stat, min_stat);
+    SSH_LOG(SSH_LOG_DEBUG, "acquiring credentials %d, %d", maj_stat, min_stat);
 
     /* finding which OID from client we selected */
     for (i=0 ; i< n_oid ; ++i){
@@ -308,7 +282,7 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
             continue;
         }
         if(len < 2 || oid_s[0] != SSH_OID_TAG || ((size_t)oid_s[1]) != len - 2){
-            SSH_LOG(SSH_LOG_WARNING,"GSSAPI: received invalid OID");
+            SSH_LOG(SSH_LOG_TRACE,"GSSAPI: received invalid OID");
             continue;
         }
         oid.elements = &oid_s[2];
@@ -323,6 +297,7 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
     session->gssapi->mech.elements = malloc(oid.length);
     if (session->gssapi->mech.elements == NULL){
         ssh_set_error_oom(session);
+        gss_release_oid_set(&min_stat, &selected);
         return SSH_ERROR;
     }
     memcpy(session->gssapi->mech.elements, oid.elements, oid.length);
@@ -333,18 +308,20 @@ ssh_gssapi_handle_userauth(ssh_session session, const char *user,
     return ssh_gssapi_send_response(session, oids[i]);
 }
 
-static char *ssh_gssapi_name_to_char(gss_name_t name)
+char *
+ssh_gssapi_name_to_char(gss_name_t name)
 {
     gss_buffer_desc buffer;
     OM_uint32 maj_stat, min_stat;
     char *ptr;
     maj_stat = gss_display_name(&min_stat, name, &buffer, NULL);
-    ssh_gssapi_log_error(SSH_LOG_WARNING,
+    ssh_gssapi_log_error(SSH_LOG_DEBUG,
                          "converting name",
                          maj_stat,
                          min_stat);
     ptr = malloc(buffer.length + 1);
     if (ptr == NULL) {
+        gss_release_buffer(&min_stat, &buffer);
         return NULL;
     }
     memcpy(ptr, buffer.value, buffer.length);
@@ -385,8 +362,6 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_server){
                 token, &out_token, session->server_callbacks->userdata);
         if (rc == SSH_ERROR){
             ssh_auth_reply_default(session, 0);
-            ssh_gssapi_free(session);
-            session->gssapi=NULL;
             return SSH_PACKET_USED;
         }
         if (ssh_string_len(out_token) != 0){
@@ -400,9 +375,8 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_server){
             }
             ssh_packet_send(session);
             SSH_STRING_FREE(out_token);
-        } else {
-            session->gssapi->state = SSH_GSSAPI_STATE_RCV_MIC;
         }
+        session->gssapi->state = SSH_GSSAPI_STATE_RCV_MIC;
         return SSH_PACKET_USED;
     }
     hexa = ssh_get_hexa(ssh_string_data(token),ssh_string_len(token));
@@ -414,7 +388,7 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_server){
     maj_stat = gss_accept_sec_context(&min_stat, &session->gssapi->ctx, session->gssapi->server_creds,
             &input_token, input_bindings, &client_name, NULL /*mech_oid*/, &output_token, &ret_flags,
             NULL /*time*/, &session->gssapi->client_creds);
-    ssh_gssapi_log_error(SSH_LOG_PROTOCOL,
+    ssh_gssapi_log_error(SSH_LOG_DEBUG,
                          "accepting token",
                          maj_stat,
                          min_stat);
@@ -424,13 +398,12 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_server){
         session->gssapi->canonic_user = ssh_gssapi_name_to_char(client_name);
     }
     if (GSS_ERROR(maj_stat)){
-        ssh_gssapi_log_error(SSH_LOG_WARNING,
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "Gssapi error",
                              maj_stat,
                              min_stat);
+        gss_release_buffer(&min_stat, &output_token);
         ssh_auth_reply_default(session,0);
-        ssh_gssapi_free(session);
-        session->gssapi=NULL;
         return SSH_PACKET_USED;
     }
 
@@ -438,13 +411,22 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_server){
         hexa = ssh_get_hexa(output_token.value, output_token.length);
         SSH_LOG(SSH_LOG_PACKET, "GSSAPI: sending token %s",hexa);
         SAFE_FREE(hexa);
-        ssh_buffer_pack(session->out_buffer,
-                        "bdP",
-                        SSH2_MSG_USERAUTH_GSSAPI_TOKEN,
-                        output_token.length,
-                        (size_t)output_token.length, output_token.value);
+        rc = ssh_buffer_pack(session->out_buffer,
+                             "bdP",
+                             SSH2_MSG_USERAUTH_GSSAPI_TOKEN,
+                             output_token.length,
+                             (size_t)output_token.length, output_token.value);
+        if (rc != SSH_OK) {
+            ssh_set_error_oom(session);
+            ssh_auth_reply_default(session, 0);
+            return SSH_PACKET_USED;
+        }
         ssh_packet_send(session);
     }
+
+    gss_release_buffer(&min_stat, &output_token);
+    gss_release_name(&min_stat, &client_name);
+
     if(maj_stat == GSS_S_COMPLETE){
         session->gssapi->state = SSH_GSSAPI_STATE_RCV_MIC;
     }
@@ -531,7 +513,7 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_mic)
         mic_token_buf.value = ssh_string_data(mic_token);
 
         maj_stat = gss_verify_mic(&min_stat, session->gssapi->ctx, &mic_buf, &mic_token_buf, NULL);
-        ssh_gssapi_log_error(SSH_LOG_PROTOCOL,
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "verifying MIC",
                              maj_stat,
                              min_stat);
@@ -562,7 +544,6 @@ error:
     ssh_auth_reply_default(session,0);
 
 end:
-    ssh_gssapi_free(session);
     if (mic_buffer != NULL) {
         SSH_BUFFER_FREE(mic_buffer);
     }
@@ -593,6 +574,7 @@ ssh_gssapi_creds ssh_gssapi_get_creds(ssh_session session)
  * Unlike ssh_gssapi_get_creds() this is called on the client side of an ssh
  * connection.
  *
+ * @param[in] session The session
  * @param[in] creds gssapi credentials handle.
  */
 void ssh_gssapi_set_creds(ssh_session session, const ssh_gssapi_creds creds)
@@ -648,7 +630,7 @@ fail:
 static int ssh_gssapi_match(ssh_session session, gss_OID_set *valid_oids)
 {
     OM_uint32 maj_stat, min_stat, lifetime;
-    gss_OID_set actual_mechs;
+    gss_OID_set actual_mechs = GSS_C_NO_OID_SET;
     gss_buffer_desc namebuf;
     gss_name_t client_id = GSS_C_NO_NAME;
     gss_OID oid;
@@ -674,6 +656,10 @@ static int ssh_gssapi_match(ssh_session session, gss_OID_set *valid_oids)
                                     &session->gssapi->client.creds,
                                     &actual_mechs, NULL);
         if (GSS_ERROR(maj_stat)) {
+            ssh_gssapi_log_error(SSH_LOG_WARN,
+                                 "acquiring credential",
+                                 maj_stat,
+                                 min_stat);
             ret = SSH_ERROR;
             goto end;
         }
@@ -710,6 +696,7 @@ static int ssh_gssapi_match(ssh_session session, gss_OID_set *valid_oids)
     ret = SSH_OK;
 
 end:
+    gss_release_oid_set(&min_stat, &actual_mechs);
     gss_release_name(&min_stat, &client_id);
     return ret;
 }
@@ -724,7 +711,7 @@ end:
 int ssh_gssapi_auth_mic(ssh_session session)
 {
     size_t i;
-    gss_OID_set selected; /* oid selected for authentication */
+    gss_OID_set selected = GSS_C_NO_OID_SET; /* oid selected for authentication */
     ssh_string *oids = NULL;
     int rc;
     size_t n_oids = 0;
@@ -733,6 +720,8 @@ int ssh_gssapi_auth_mic(ssh_session session)
     gss_buffer_desc hostname;
     const char *gss_host = session->opts.host;
 
+    /* Destroy earlier GSSAPI context if any */
+    ssh_gssapi_free(session);
     rc = ssh_gssapi_init(session);
     if (rc == SSH_ERROR) {
         return SSH_AUTH_ERROR;
@@ -750,8 +739,8 @@ int ssh_gssapi_auth_mic(ssh_session session)
                                (gss_OID)GSS_C_NT_HOSTBASED_SERVICE,
                                &session->gssapi->client.server_name);
     if (maj_stat != GSS_S_COMPLETE) {
-        SSH_LOG(SSH_LOG_WARNING, "importing name %d, %d", maj_stat, min_stat);
-        ssh_gssapi_log_error(SSH_LOG_WARNING,
+        SSH_LOG(SSH_LOG_DEBUG, "importing name %d, %d", maj_stat, min_stat);
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "importing name",
                              maj_stat,
                              min_stat);
@@ -765,7 +754,7 @@ int ssh_gssapi_auth_mic(ssh_session session)
         return SSH_AUTH_ERROR;
     }
 
-    SSH_LOG(SSH_LOG_PROTOCOL, "Authenticating with gssapi to host %s with user %s",
+    SSH_LOG(SSH_LOG_DEBUG, "Authenticating with gssapi to host %s with user %s",
             session->opts.host, session->gssapi->user);
     rc = ssh_gssapi_match(session, &selected);
     if (rc == SSH_ERROR) {
@@ -773,7 +762,7 @@ int ssh_gssapi_auth_mic(ssh_session session)
     }
 
     n_oids = selected->count;
-    SSH_LOG(SSH_LOG_PROTOCOL, "Sending %zu oids", n_oids);
+    SSH_LOG(SSH_LOG_DEBUG, "Sending %zu oids", n_oids);
 
     oids = calloc(n_oids, sizeof(ssh_string));
     if (oids == NULL) {
@@ -801,6 +790,8 @@ out:
         SSH_STRING_FREE(oids[i]);
     }
     free(oids);
+    gss_release_oid_set(&min_stat, &selected);
+
     if (rc != SSH_ERROR) {
         return SSH_AUTH_AGAIN;
     }
@@ -845,6 +836,7 @@ static gss_OID ssh_gssapi_oid_from_string(ssh_string oid_s)
 }
 
 SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_response){
+    int rc;
     ssh_string oid_s;
     gss_uint32 maj_stat, min_stat;
     gss_buffer_desc input_token = GSS_C_EMPTY_BUFFER;
@@ -886,7 +878,7 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_response){
                                     0, NULL, &input_token, NULL,
                                     &output_token, NULL, NULL);
     if(GSS_ERROR(maj_stat)){
-        ssh_gssapi_log_error(SSH_LOG_WARNING,
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "Initializing gssapi context",
                              maj_stat,
                              min_stat);
@@ -896,20 +888,24 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_response){
         hexa = ssh_get_hexa(output_token.value, output_token.length);
         SSH_LOG(SSH_LOG_PACKET, "GSSAPI: sending token %s", hexa);
         SAFE_FREE(hexa);
-        ssh_buffer_pack(session->out_buffer,
-                        "bdP",
-                        SSH2_MSG_USERAUTH_GSSAPI_TOKEN,
-                        output_token.length,
-                        (size_t)output_token.length, output_token.value);
+        rc = ssh_buffer_pack(session->out_buffer,
+                             "bdP",
+                             SSH2_MSG_USERAUTH_GSSAPI_TOKEN,
+                             output_token.length,
+                             (size_t)output_token.length, output_token.value);
+        if (rc != SSH_OK) {
+            ssh_set_error_oom(session);
+            goto error;
+        }
         ssh_packet_send(session);
         session->auth.state = SSH_AUTH_STATE_GSSAPI_TOKEN;
     }
+
+    gss_release_buffer(&min_stat, &output_token);
     return SSH_PACKET_USED;
 
 error:
     session->auth.state = SSH_AUTH_STATE_ERROR;
-    ssh_gssapi_free(session);
-    session->gssapi = NULL;
     return SSH_PACKET_USED;
 }
 
@@ -933,9 +929,11 @@ static int ssh_gssapi_send_mic(ssh_session session)
 
     maj_stat = gss_get_mic(&min_stat,session->gssapi->ctx, GSS_C_QOP_DEFAULT,
                            &mic_buf, &mic_token_buf);
+
+    SSH_BUFFER_FREE(mic_buffer);
+
     if (GSS_ERROR(maj_stat)){
-        SSH_BUFFER_FREE(mic_buffer);
-        ssh_gssapi_log_error(SSH_LOG_PROTOCOL,
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "generating MIC",
                              maj_stat,
                              min_stat);
@@ -947,8 +945,10 @@ static int ssh_gssapi_send_mic(ssh_session session)
                          SSH2_MSG_USERAUTH_GSSAPI_MIC,
                          mic_token_buf.length,
                          (size_t)mic_token_buf.length, mic_token_buf.value);
+
+    gss_release_buffer(&min_stat, &mic_token_buf);
+
     if (rc != SSH_OK) {
-        SSH_BUFFER_FREE(mic_buffer);
         ssh_set_error_oom(session);
         return SSH_ERROR;
     }
@@ -957,6 +957,7 @@ static int ssh_gssapi_send_mic(ssh_session session)
 }
 
 SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_client){
+    int rc;
     ssh_string token;
     char *hexa;
     OM_uint32 maj_stat, min_stat;
@@ -992,13 +993,13 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_client){
                                     0, NULL, &input_token, NULL,
                                     &output_token, NULL, NULL);
 
-    ssh_gssapi_log_error(SSH_LOG_PROTOCOL,
+    ssh_gssapi_log_error(SSH_LOG_DEBUG,
                          "accepting token",
                          maj_stat,
                          min_stat);
     SSH_STRING_FREE(token);
     if (GSS_ERROR(maj_stat)){
-        ssh_gssapi_log_error(SSH_LOG_PROTOCOL,
+        ssh_gssapi_log_error(SSH_LOG_DEBUG,
                              "Gssapi error",
                              maj_stat,
                              min_stat);
@@ -1009,14 +1010,20 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_client){
         hexa = ssh_get_hexa(output_token.value, output_token.length);
         SSH_LOG(SSH_LOG_PACKET, "GSSAPI: sending token %s",hexa);
         SAFE_FREE(hexa);
-        ssh_buffer_pack(session->out_buffer,
-                        "bdP",
-                        SSH2_MSG_USERAUTH_GSSAPI_TOKEN,
-                        output_token.length,
-                        (size_t)output_token.length, output_token.value);
+        rc = ssh_buffer_pack(session->out_buffer,
+                             "bdP",
+                             SSH2_MSG_USERAUTH_GSSAPI_TOKEN,
+                             output_token.length,
+                             (size_t)output_token.length, output_token.value);
+        if (rc != SSH_OK) {
+            ssh_set_error_oom(session);
+            goto error;
+        }
         ssh_packet_send(session);
     }
 
+    gss_release_buffer(&min_stat, &output_token);
+
     if (maj_stat == GSS_S_COMPLETE) {
         ssh_gssapi_send_mic(session);
         session->auth.state = SSH_AUTH_STATE_GSSAPI_MIC_SENT;
@@ -1026,7 +1033,5 @@ SSH_PACKET_CALLBACK(ssh_packet_userauth_gssapi_token_client){
 
 error:
     session->auth.state = SSH_AUTH_STATE_ERROR;
-    ssh_gssapi_free(session);
-    session->gssapi = NULL;
     return SSH_PACKET_USED;
 }
