@@ -66,7 +66,7 @@ ssh_scp ssh_scp_new(ssh_session session, int mode, const char *location)
 {
     ssh_scp scp = NULL;
 
-    if (session == NULL) {
+    if (session == NULL || location == NULL) {
         goto error;
     }
 
@@ -146,7 +146,7 @@ int ssh_scp_init(ssh_scp scp)
         return SSH_ERROR;
     }
 
-    SSH_LOG(SSH_LOG_PROTOCOL, "Initializing scp session %s %son location '%s'",
+    SSH_LOG(SSH_LOG_DEBUG, "Initializing scp session %s %son location '%s'",
             scp->mode == SSH_SCP_WRITE?"write":"read",
             scp->recursive ? "recursive " : "",
             scp->location);
@@ -266,7 +266,7 @@ int ssh_scp_close(ssh_scp scp)
          */
         while (!ssh_channel_is_eof(scp->channel)) {
             rc = ssh_channel_read(scp->channel, buffer, sizeof(buffer), 0);
-            if (rc == SSH_ERROR || rc == 0) {
+            if (rc == SSH_ERROR || rc == SSH_AGAIN || rc == 0) {
                 break;
             }
         }
@@ -376,7 +376,7 @@ int ssh_scp_push_directory(ssh_scp scp, const char *dirname, int mode)
         goto error;
     }
 
-    SSH_LOG(SSH_LOG_PROTOCOL,
+    SSH_LOG(SSH_LOG_DEBUG,
             "SCP pushing directory %s with permissions '%s'",
             vis_encoded, perms);
 
@@ -517,7 +517,7 @@ int ssh_scp_push_file64(ssh_scp scp, const char *filename, uint64_t size,
         goto error;
     }
 
-    SSH_LOG(SSH_LOG_PROTOCOL,
+    SSH_LOG(SSH_LOG_DEBUG,
             "SCP pushing file %s, size %" PRIu64 " with permissions '%s'",
             vis_encoded, size, perms);
 
@@ -603,6 +603,12 @@ int ssh_scp_response(ssh_scp scp, char **response)
 
     rc = ssh_channel_read(scp->channel, &code, 1, 0);
     if (rc == SSH_ERROR) {
+        scp->state = SSH_SCP_ERROR;
+        return SSH_ERROR;
+    }
+    if (rc == SSH_AGAIN) {
+        ssh_set_error(scp->session, SSH_FATAL, "SCP: ssh_channel_read timeout");
+        scp->state = SSH_SCP_ERROR;
         return SSH_ERROR;
     }
 
@@ -760,6 +766,14 @@ int ssh_scp_read_string(ssh_scp scp, char *buffer, size_t len)
             break;
         }
 
+        if (err == SSH_AGAIN) {
+            ssh_set_error(scp->session,
+                          SSH_FATAL,
+                          "SCP: ssh_channel_read timeout");
+            err = SSH_ERROR;
+            break;
+        }
+
         read++;
         if (buffer[read - 1] == '\n') {
             break;
@@ -825,7 +839,7 @@ int ssh_scp_pull_request(ssh_scp scp)
         *p = '\0';
     }
 
-    SSH_LOG(SSH_LOG_PROTOCOL, "Received SCP request: '%s'", buffer);
+    SSH_LOG(SSH_LOG_DEBUG, "Received SCP request: '%s'", buffer);
     switch(buffer[0]) {
     case 'C':
         /* File */
@@ -1027,12 +1041,16 @@ int ssh_scp_read(ssh_scp scp, void *buffer, size_t size)
     }
 
     rc = ssh_channel_read(scp->channel, buffer, size, 0);
-    if (rc != SSH_ERROR) {
-        scp->processed += rc;
-    } else {
+    if (rc == SSH_ERROR) {
         scp->state = SSH_SCP_ERROR;
         return SSH_ERROR;
     }
+    if (rc == SSH_AGAIN) {
+        ssh_set_error(scp->session, SSH_FATAL, "SCP: ssh_channel_read timeout");
+        scp->state = SSH_SCP_ERROR;
+        return SSH_ERROR;
+    }
+    scp->processed += rc;
 
     /* Check if we arrived at end of file */
     if (scp->processed == scp->filelen) {
