@@ -1,5 +1,5 @@
 /*
- * wrapper.c - wrapper for crytpo functions
+ * wrapper.c - wrapper for crypto functions
  *
  * This file is part of the SSH Library
  *
@@ -150,15 +150,15 @@ static void cipher_free(struct ssh_cipher_struct *cipher) {
   SAFE_FREE(cipher);
 }
 
-struct ssh_crypto_struct *crypto_new(void) {
-   struct ssh_crypto_struct *crypto;
+struct ssh_crypto_struct *crypto_new(void)
+{
+    struct ssh_crypto_struct *crypto;
 
-  crypto = malloc(sizeof(struct ssh_crypto_struct));
-  if (crypto == NULL) {
-    return NULL;
-  }
-  ZERO_STRUCTP(crypto);
-  return crypto;
+    crypto = calloc(1, sizeof(struct ssh_crypto_struct));
+    if (crypto == NULL) {
+        return NULL;
+    }
+    return crypto;
 }
 
 void crypto_free(struct ssh_crypto_struct *crypto)
@@ -176,13 +176,9 @@ void crypto_free(struct ssh_crypto_struct *crypto)
 #ifdef HAVE_ECDH
     SAFE_FREE(crypto->ecdh_client_pubkey);
     SAFE_FREE(crypto->ecdh_server_pubkey);
-    if(crypto->ecdh_privkey != NULL){
+    if (crypto->ecdh_privkey != NULL) {
 #ifdef HAVE_OPENSSL_ECC
-/* TODO Change to new API when the OpenSSL will support export of uncompressed EC keys
- * https://github.com/openssl/openssl/pull/16624
- * #if OPENSSL_VERSION_NUMBER < 0x30000000L
- */
-#if 1
+#if OPENSSL_VERSION_NUMBER < 0x30000000L
         EC_KEY_free(crypto->ecdh_privkey);
 #else
         EVP_PKEY_free(crypto->ecdh_privkey);
@@ -203,14 +199,12 @@ void crypto_free(struct ssh_crypto_struct *crypto)
         SAFE_FREE(crypto->secret_hash);
     }
 #ifdef WITH_ZLIB
-    if (crypto->compress_out_ctx &&
-        (deflateEnd(crypto->compress_out_ctx) != 0)) {
-        inflateEnd(crypto->compress_out_ctx);
+    if (crypto->compress_out_ctx) {
+        deflateEnd(crypto->compress_out_ctx);
     }
     SAFE_FREE(crypto->compress_out_ctx);
 
-    if (crypto->compress_in_ctx &&
-        (deflateEnd(crypto->compress_in_ctx) != 0)) {
+    if (crypto->compress_in_ctx) {
         inflateEnd(crypto->compress_in_ctx);
     }
     SAFE_FREE(crypto->compress_in_ctx);
@@ -242,9 +236,32 @@ void crypto_free(struct ssh_crypto_struct *crypto)
     SAFE_FREE(crypto);
 }
 
+static void
+compression_enable(ssh_session session,
+                   enum ssh_crypto_direction_e direction,
+                   bool delayed)
+{
+    /* The delayed compression is turned on AFTER authentication. This means
+     * that we need to turn it on immediately in case of rekeying */
+    if (delayed && !(session->flags & SSH_SESSION_FLAG_AUTHENTICATED)) {
+        if (direction == SSH_DIRECTION_IN) {
+            session->next_crypto->delayed_compress_in = 1;
+        } else { /* SSH_DIRECTION_OUT */
+            session->next_crypto->delayed_compress_out = 1;
+        }
+    } else {
+        if (direction == SSH_DIRECTION_IN) {
+            session->next_crypto->do_compress_in = 1;
+        } else { /* SSH_DIRECTION_OUT */
+            session->next_crypto->do_compress_out = 1;
+        }
+    }
+}
+
 static int crypt_set_algorithms2(ssh_session session)
 {
     const char *wanted = NULL;
+    const char *method = NULL;
     struct ssh_cipher_struct *ssh_ciphertab=ssh_get_ciphertab();
     struct ssh_hmac_struct *ssh_hmactab=ssh_get_hmactab();
     uint8_t i = 0;
@@ -367,22 +384,29 @@ static int crypt_set_algorithms2(ssh_session session)
     session->next_crypto->in_hmac = ssh_hmactab[i].hmac_type;
     session->next_crypto->in_hmac_etm = ssh_hmactab[i].etm;
 
-    /* compression */
-    cmp = strcmp(session->next_crypto->kex_methods[SSH_COMP_C_S], "zlib");
+    /* compression: client */
+    method = session->next_crypto->kex_methods[SSH_COMP_C_S];
+    cmp = strcmp(method, "zlib");
     if (cmp == 0) {
-        session->next_crypto->do_compress_out = 1;
+        SSH_LOG(SSH_LOG_PACKET, "enabling C->S compression");
+        compression_enable(session, SSH_DIRECTION_OUT, false);
     }
-    cmp = strcmp(session->next_crypto->kex_methods[SSH_COMP_S_C], "zlib");
+    cmp = strcmp(method, "zlib@openssh.com");
     if (cmp == 0) {
-        session->next_crypto->do_compress_in = 1;
+        SSH_LOG(SSH_LOG_PACKET, "enabling C->S delayed compression");
+        compression_enable(session, SSH_DIRECTION_OUT, true);
     }
-    cmp = strcmp(session->next_crypto->kex_methods[SSH_COMP_C_S], "zlib@openssh.com");
+
+    method = session->next_crypto->kex_methods[SSH_COMP_S_C];
+    cmp = strcmp(method, "zlib");
     if (cmp == 0) {
-        session->next_crypto->delayed_compress_out = 1;
+        SSH_LOG(SSH_LOG_PACKET, "enabling S->C compression");
+        compression_enable(session, SSH_DIRECTION_IN, false);
     }
-    cmp = strcmp(session->next_crypto->kex_methods[SSH_COMP_S_C], "zlib@openssh.com");
+    cmp = strcmp(method, "zlib@openssh.com");
     if (cmp == 0) {
-        session->next_crypto->delayed_compress_in = 1;
+        SSH_LOG(SSH_LOG_PACKET, "enabling S->C delayed compression");
+        compression_enable(session, SSH_DIRECTION_IN, true);
     }
 
     return SSH_OK;
@@ -401,7 +425,6 @@ int crypt_set_algorithms_server(ssh_session session){
     struct ssh_hmac_struct   *ssh_hmactab=ssh_get_hmactab();
     int cmp;
 
-
     if (session == NULL) {
         return SSH_ERROR;
     }
@@ -520,33 +543,27 @@ int crypt_set_algorithms_server(ssh_session session){
 
     /* compression */
     method = session->next_crypto->kex_methods[SSH_COMP_C_S];
-    if(strcmp(method,"zlib") == 0){
-        SSH_LOG(SSH_LOG_PACKET,"enabling C->S compression");
-        session->next_crypto->do_compress_in=1;
+    cmp = strcmp(method, "zlib");
+    if (cmp == 0) {
+        SSH_LOG(SSH_LOG_PACKET, "enabling C->S compression");
+        compression_enable(session, SSH_DIRECTION_IN, false);
     }
-    if(strcmp(method,"zlib@openssh.com") == 0){
-        SSH_LOG(SSH_LOG_PACKET,"enabling C->S delayed compression");
-
-        if (session->flags & SSH_SESSION_FLAG_AUTHENTICATED) {
-            session->next_crypto->do_compress_in = 1;
-        } else {
-            session->next_crypto->delayed_compress_in = 1;
-        }
+    cmp = strcmp(method, "zlib@openssh.com");
+    if (cmp == 0) {
+        SSH_LOG(SSH_LOG_PACKET, "enabling C->S delayed compression");
+        compression_enable(session, SSH_DIRECTION_IN, true);
     }
 
     method = session->next_crypto->kex_methods[SSH_COMP_S_C];
-    if(strcmp(method,"zlib") == 0){
+    cmp = strcmp(method, "zlib");
+    if (cmp == 0) {
         SSH_LOG(SSH_LOG_PACKET, "enabling S->C compression");
-        session->next_crypto->do_compress_out=1;
+        compression_enable(session, SSH_DIRECTION_OUT, false);
     }
-    if(strcmp(method,"zlib@openssh.com") == 0){
-        SSH_LOG(SSH_LOG_PACKET,"enabling S->C delayed compression");
-
-        if (session->flags & SSH_SESSION_FLAG_AUTHENTICATED) {
-            session->next_crypto->do_compress_out = 1;
-        } else {
-            session->next_crypto->delayed_compress_out = 1;
-        }
+    cmp = strcmp(method, "zlib@openssh.com");
+    if (cmp == 0) {
+        SSH_LOG(SSH_LOG_PACKET, "enabling S->C delayed compression");
+        compression_enable(session, SSH_DIRECTION_OUT, true);
     }
 
     method = session->next_crypto->kex_methods[SSH_HOSTKEYS];
