@@ -218,6 +218,39 @@ static void torture_max_sessions(void **state)
 #undef MAX_CHANNELS
 }
 
+static void torture_no_more_sessions(void **state)
+{
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    ssh_channel channels[2];
+    int rc;
+
+    /* Open a channel session */
+    channels[0] = ssh_channel_new(session);
+    assert_non_null(channels[0]);
+
+    rc = ssh_channel_open_session(channels[0]);
+    assert_ssh_return_code(session, rc);
+
+    /* Send no-more-sessions@openssh.com global request */
+    rc = ssh_request_no_more_sessions(session);
+    assert_ssh_return_code(session, rc);
+
+    /* Try to open an extra session and expect failure */
+    channels[1] = ssh_channel_new(session);
+    assert_non_null(channels[1]);
+
+    rc = ssh_channel_open_session(channels[1]);
+    assert_int_equal(rc, SSH_ERROR);
+
+    /* Free the unused channel */
+    ssh_channel_free(channels[1]);
+
+    /* Close and free open channel */
+    ssh_channel_close(channels[0]);
+    ssh_channel_free(channels[0]);
+}
+
 static void torture_channel_delayed_close(void **state)
 {
     struct torture_state *s = *state;
@@ -295,7 +328,7 @@ static void torture_freed_channel_poll_timeout(void **state)
     struct torture_state *s = *state;
     ssh_session session = s->ssh.session;
     ssh_channel channel;
-
+    bool channel_freed = false;
     char request[256];
     char buff[256] = {0};
     int rc;
@@ -318,10 +351,19 @@ static void torture_freed_channel_poll_timeout(void **state)
     } while(rc > 0);
     assert_ssh_return_code(session, rc);
 
+    /* when either of these conditions is met the call to ssh_channel_free will
+     * actually free the channel so calling poll on that channel will be
+     * use-after-free */
+    if ((channel->flags & SSH_CHANNEL_FLAG_CLOSED_REMOTE) ||
+        (channel->flags & SSH_CHANNEL_FLAG_NOT_BOUND)) {
+        channel_freed = true;
+    }
     ssh_channel_free(channel);
 
-    rc = ssh_channel_poll_timeout(channel, 500, 0);
-    assert_int_equal(rc, SSH_ERROR);
+    if (!channel_freed) {
+        rc = ssh_channel_poll_timeout(channel, 500, 0);
+        assert_int_equal(rc, SSH_ERROR);
+    }
 }
 
 /* Ensure that calling 'ssh_channel_read_nonblocking' on a freed channel does
@@ -355,6 +397,70 @@ static void torture_freed_channel_read_nonblocking(void **state)
     assert_ssh_return_code_equal(session, rc, SSH_ERROR);
 }
 
+static void torture_channel_exit_status(void **state)
+{
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    ssh_channel channel = NULL;
+    char request[256];
+    uint32_t exit_status = (uint32_t)-1;
+    int rc;
+
+    rc = snprintf(request, sizeof(request), "true");
+    assert_return_code(rc, errno);
+
+    channel = ssh_channel_new(session);
+    assert_non_null(channel);
+
+    rc = ssh_channel_open_session(channel);
+    assert_ssh_return_code(session, rc);
+
+    /* Make the request, read parts with close */
+    rc = ssh_channel_request_exec(channel, request);
+    assert_ssh_return_code(session, rc);
+
+    exit_status = ssh_channel_get_exit_state(channel, &exit_status, NULL, NULL);
+    assert_ssh_return_code(session, rc);
+    assert_int_equal(exit_status, 0);
+}
+
+static void torture_channel_exit_signal(void **state)
+{
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    ssh_channel channel = NULL;
+    char request[256];
+    uint32_t exit_status = (uint32_t)-1;
+    char *exit_signal = NULL;
+    int core_dumped = false;
+    int rc;
+
+    rc = snprintf(request, sizeof(request), "cat");
+    assert_return_code(rc, errno);
+
+    channel = ssh_channel_new(session);
+    assert_non_null(channel);
+
+    rc = ssh_channel_open_session(channel);
+    assert_ssh_return_code(session, rc);
+
+    /* Make the request, read parts with close */
+    rc = ssh_channel_request_exec(channel, request);
+    assert_ssh_return_code(session, rc);
+    rc = ssh_channel_request_send_signal(channel, "TERM");
+    assert_ssh_return_code(session, rc);
+
+    exit_status = ssh_channel_get_exit_state(channel,
+                                             &exit_status,
+                                             &exit_signal,
+                                             &core_dumped);
+    assert_ssh_return_code(session, rc);
+    assert_int_equal(exit_status, 0);
+    assert_string_equal(exit_signal, "TERM");
+    SAFE_FREE(exit_signal);
+}
+
+
 /* Ensure that calling 'ssh_channel_get_exit_status' on a freed channel does not
  * lead to segmentation faults. */
 static void torture_freed_channel_get_exit_status(void **state)
@@ -362,7 +468,7 @@ static void torture_freed_channel_get_exit_status(void **state)
     struct torture_state *s = *state;
     ssh_session session = s->ssh.session;
     ssh_channel channel;
-
+    bool channel_freed = false;
     char request[256];
     char buff[256] = {0};
     int rc;
@@ -385,10 +491,55 @@ static void torture_freed_channel_get_exit_status(void **state)
     } while(rc > 0);
     assert_ssh_return_code(session, rc);
 
-    ssh_channel_free(channel);
+    /* when either of these conditions is met the call to ssh_channel_free will
+     * actually free the channel so calling poll on that channel will be
+     * use-after-free */
+    if ((channel->flags & SSH_CHANNEL_FLAG_CLOSED_REMOTE) ||
+        (channel->flags & SSH_CHANNEL_FLAG_NOT_BOUND)) {
+        channel_freed = true;
+    }
+    SSH_CHANNEL_FREE(channel);
 
-    rc = ssh_channel_get_exit_status(channel);
-    assert_ssh_return_code_equal(session, rc, SSH_ERROR);
+    if (!channel_freed) {
+        rc = ssh_channel_get_exit_status(channel);
+        assert_ssh_return_code_equal(session, rc, SSH_ERROR);
+    }
+}
+
+static void
+torture_channel_read_stderr(void **state)
+{
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    ssh_channel channel;
+    int rc;
+
+    channel = ssh_channel_new(session);
+    assert_non_null(channel);
+
+    rc = ssh_channel_open_session(channel);
+    assert_ssh_return_code(session, rc);
+
+    /* This writes to standard error "pipe" */
+    rc = ssh_channel_request_exec(channel, "echo -n ABCD >&2");
+    assert_ssh_return_code(session, rc);
+
+    /* No data in stdout */
+    rc = ssh_channel_read(channel, buffer, sizeof(buffer), 0);
+    assert_int_equal(rc, 0);
+
+    /* poll should say how much we can read */
+    rc = ssh_channel_poll(channel, 1);
+    assert_int_equal(rc, strlen("ABCD"));
+
+    /* Everything in stderr */
+    rc = ssh_channel_read(channel, buffer, sizeof(buffer), 1);
+    assert_int_equal(rc, strlen("ABCD"));
+
+    buffer[rc] = '\0';
+    assert_string_equal("ABCD", buffer);
+
+    ssh_channel_free(channel);
 }
 
 static void torture_pubkey_hash(void **state)
@@ -438,6 +589,9 @@ int torture_run_tests(void) {
         cmocka_unit_test_setup_teardown(torture_max_sessions,
                                         session_setup,
                                         session_teardown),
+        cmocka_unit_test_setup_teardown(torture_no_more_sessions,
+                                        session_setup,
+                                        session_teardown),
         cmocka_unit_test_setup_teardown(torture_channel_delayed_close,
                                         session_setup,
                                         session_teardown),
@@ -450,9 +604,18 @@ int torture_run_tests(void) {
         cmocka_unit_test_setup_teardown(torture_freed_channel_read_nonblocking,
                                         session_setup,
                                         session_teardown),
+        cmocka_unit_test_setup_teardown(torture_channel_exit_status,
+                                        session_setup,
+                                        session_teardown),
+        cmocka_unit_test_setup_teardown(torture_channel_exit_signal,
+                                        session_setup,
+                                        session_teardown),
         cmocka_unit_test_setup_teardown(torture_freed_channel_get_exit_status,
                                         session_setup,
                                         session_teardown),
+        cmocka_unit_test_setup_teardown(torture_channel_read_stderr,
+                                        session_setup,
+                                        session_teardown),
         cmocka_unit_test_setup_teardown(torture_pubkey_hash,
                                         session_setup,
                                         session_teardown),
